import asyncio
from contextlib import suppress
from typing import (
    TYPE_CHECKING,
    Tuple
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.core import funcs
from xoa_driver.internals.core.exceptions import XmpBadValueError
from xoa_driver.internals import exceptions
from xoa_driver.internals.commands import (
    C_LOGON,
    C_LOGOFF,
    C_OWNER,
    C_TIMEOUT,
    C_KEEPALIVE,
    C_INDICES,
    C_STATSESSION,
)


class TesterSession:
    """
    Establishing communication session through connection with the tester.

    :param conn: connection handler
    :type host: itf.IConnection
    :param owner_name: username of the user
    :type username: str
    :param password: login password of the tester, defaults to "xena"
    :type password: str, optional
    :param timeout_seconds: the timeout to close idle connection
    :type port: int, optional
    :param keepalive: `True` if keepalive signal is needed, and `False` otherwise, defaults to `False`.
    :type keepalive: bool, optional
    """

    __slots__ = ("_conn", "owner_name", "pwd", "timeout", "keepalive")

    def __init__(self, conn: "itf.IConnection", owner_name: str, password: str = "xena", timeout_seconds: int = 130, keepalive: bool = False) -> None:
        self._conn = conn
        self.owner_name = owner_name
        self.pwd = password
        self.timeout = timeout_seconds
        self.keepalive = keepalive

    async def logon(self) -> "TesterSession":
        logon, *_ = await funcs.apply(
            C_LOGON(self._conn).set(self.pwd),
            C_OWNER(self._conn).set(self.owner_name),
            C_TIMEOUT(self._conn).set(self.timeout),
            return_exceptions=True,
            token_timeout_sec=1.0
        )
        if isinstance(logon, XmpBadValueError):
            raise exceptions.WrongTesterPasswordError(self.pwd) from None

        if self.keepalive:
            asyncio.create_task(self.__do_keepalive(), name="keepalive").add_done_callback(
                self.__handle_exceptions
            )
        return self

    async def chang_timeout(self, seconds: int = 130) -> None:
        """Modify session timeout, """
        if not self.is_online:
            raise RuntimeError("Timeout can be changet only after connection is established.")
        self.timeout = seconds
        await C_TIMEOUT(self._conn).set(self.timeout)

    def __handle_exceptions(self, fut: asyncio.Future) -> None:
        if fut.cancelled():
            return None
        elif e := fut.exception():
            raise e

    async def __do_keepalive(self) -> None:
        while self._conn.is_connected:
            with suppress(Exception):
                await C_KEEPALIVE(self._conn).get()
            await asyncio.sleep(max(0, self.timeout - 5))

    @property
    def is_online(self) -> bool:
        """
        Check if connection is still active.

        :return: connection is still on or not.
        :rtype: bool
        """

        return self._conn.is_connected

    async def logoff(self) -> None:
        """
        Gracefully log off from the tester.
        """

        await C_LOGOFF(self._conn).set()
        self._conn.close()

    async def sessions_info(self) -> Tuple[C_STATSESSION.GetDataAttr, ...]:
        """
        Return information about all active sessions on the tester.

        :return: a list of session information
        :rtype: List[C_STATSESSION]
        """

        sessions = await C_INDICES(self._conn).get()
        session_ids = sessions.session_ids
        query_sessions = (
            C_STATSESSION(self._conn, _session_xindex=sid).get()
            for sid in session_ids
        )
        return tuple(await funcs.apply(*query_sessions))



import functools
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    P_FLASH,
    P_SPEEDSELECTION,
    P_SPEEDS_SUPPORTED,
    P_UAT_MODE,
    P_UAT_FLR,
    P_STATUS,
    P_TCVRSTATUS,
    P_LOADMODE,
    PP_FECMODE,
)
from xoa_driver.internals.utils import attributes as utils
from xoa_driver.internals.utils.indices import index_manager as idx_mgr
from xoa_driver.internals.hli_v2.indices.streams.genuine_stream import GenuineStreamIdx
from xoa_driver.internals.hli_v2.indices.filter.genuine_filter import GenuineFilterIdx
from xoa_driver.internals.hli_v2.indices.port_dataset import PortDatasetIdx
from xoa_driver.internals.state_storage import ports_state

from .port_l23 import (
    BasePortL23,
    Speed,
    # TxConfiguration,
)

from .port_transceiver import PortTransceiver
from .port_reception_statistics import GenuinePortReceptionStatistics
from .port_transmission_statistics import PortTransmissionStatistics

StreamIndices = idx_mgr.IndexManager[GenuineStreamIdx]
FilterIndices = idx_mgr.IndexManager[GenuineFilterIdx]
PortDatasetIndices = idx_mgr.IndexManager[PortDatasetIdx]


class SpeedMode:
    """L23 port's speed mode"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.selection = P_SPEEDSELECTION(conn, module_id, port_id)
        """L23 port speed mode selection.
        Representation of P_SPEEDSELECTION
        """

        self.supported = P_SPEEDS_SUPPORTED(conn, module_id, port_id)
        """L23 port's supported speed modes.
        Representation of P_SPEEDS_SUPPORTED
        """


class GenuineSpeed(Speed):
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.mode = SpeedMode(conn, module_id, port_id)
        """L23 port's speed mode."""


class UnAvailableTime:
    """UnAvailable Time"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.mode = P_UAT_MODE(conn, module_id, port_id)
        """L23 port's Unavailable Time mode.
        Representation of P_UAT_MODE
        """

        self.frame_loss_ratio = P_UAT_FLR(conn, module_id, port_id)
        """L23 port's Frame Loss Ratio for UAT.
        Representation of P_UAT_FLR
        """


class PortStatistics:
    """L23 port statistics"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.rx = GenuinePortReceptionStatistics(conn, module_id, port_id)
        """L23 port's RX statistics."""

        self.tx = PortTransmissionStatistics(conn, module_id, port_id)
        """L23 port's TX statistics."""


class BasePortL23Genuine(BasePortL23):
    """L23 port basic configuration."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)

        self._local_states = ports_state.PortL23GenuineLocalState()

        self.flash = P_FLASH(conn, module_id, port_id)
        """L23 port flashes.
        Representation of P_FLASH
        """

        self.status = P_STATUS(conn, module_id, port_id)
        """L23 port's received optical signal level'.
        Representation of P_STATUS
        """

        self.config_load_mode = P_LOADMODE(conn, module_id, port_id)
        """Chimera Port loading mode.
        Representation of P_LOADMODE
        """

        self.tcvr_status = P_TCVRSTATUS(conn, module_id, port_id)
        """L23 port transceiver status information.
        Representation of P_TCVRSTATUS
        """

        self.fec_mode = PP_FECMODE(conn, module_id, port_id)
        """L23 port FEC mode.
        Representation of PP_FECMODE
        """

        self.speed = GenuineSpeed(conn, module_id, port_id)
        """L23 port speed configuration.
        """

        self.uat = UnAvailableTime(conn, module_id, port_id)
        """L23 port UnAvailable Time configuration.
        """

        self.transceiver = PortTransceiver(conn, module_id, port_id)
        """L23 port transceiver configuration.
        """

        self.statistics = PortStatistics(conn, module_id, port_id)
        """L23 port statistics.
        """

        self.streams: StreamIndices = idx_mgr.IndexManager(
            conn,
            GenuineStreamIdx,
            module_id,
            port_id
        )
        """L23 port stream index manager.
        """

        self.filters: FilterIndices = idx_mgr.IndexManager(
            conn,
            GenuineFilterIdx,
            module_id,
            port_id
        )
        """L23 port filter index manager.
        """

        self.datasets: PortDatasetIndices = idx_mgr.IndexManager(
            conn,
            PortDatasetIdx,
            module_id,
            port_id
        )
        """L23 port histogram index manager.
        """

    @property
    def info(self) -> ports_state.PortL23GenuineLocalState:
        return self._local_states

    async def _setup(self):
        await self._local_states.initiate(self)
        self._local_states.register_subscriptions(self)
        return self

    on_speed_selection_change = functools.partialmethod(utils.on_event, P_SPEEDSELECTION)
    """Register a callback to the event that the port's speed mode changes."""



from .field import field
from .base_struct import (
    RequestBodyStruct,
    ResponseBodyStruct,
)
from .types import (
    XmpByte,
    XmpHex,
    XmpInt,
    XmpIPv4Address,
    XmpIPv6Address,
    XmpLong,
    XmpMacAddress,
    XmpSequence,
    XmpShort,
    XmpStr,
    Hex
)

__all__ = (
    "field",
    "RequestBodyStruct",
    "ResponseBodyStruct",
    "XmpByte",
    "XmpHex",
    "XmpInt",
    "XmpIPv4Address",
    "XmpIPv6Address",
    "XmpLong",
    "XmpMacAddress",
    "XmpSequence",
    "XmpShort",
    "XmpStr",
    "Hex",
)


"""
Commands with non fixed size attributes will work slower

!!!NOT PERMITED STRUCTURES!!!


# 2 strings with undefined size in one packet, seporated or not with other attributes

class BadClass1:
    attr1: str = field(XmpStr())
    attr2: str = field(XmpStr())


# Strings and arrays with undefined size

class BadClass2:
    attr1: str = field(XmpStr())
    attr2: list[int] = field(XmpSequence(XmpInt()))


# List of strings

class BadClass3:
    attr1: list[str] = field(XmpSequence(XmpStr()))


# Multiple lists of unfixed sizes

class BadClass3:
    attr1: list[int] = field(XmpSequence(XmpInt()))
    attr2: list[int] = field(XmpSequence(XmpInt()))

"""



from typing import (
    TYPE_CHECKING,
    List,
    Type,
    TypeVar,
)
from xoa_driver.internals.commands import (
    PF_INDICES,
    PF_CREATE,
    PF_DELETE,
    PF_ENABLE,
    PF_COMMENT,
    PF_CONDITION,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.utils import kind
from xoa_driver.internals.utils.indices import observer as idx_obs
from ..base_index import BaseIndex


FT = TypeVar("FT")


class BaseFilterIdx(BaseIndex):
    """Base L23 Filter Index Manager"""
    def __init__(self, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> None:
        super().__init__(conn, kind, observer)

        self.comment = PF_COMMENT(conn, *kind)
        """Description of the filter.

        :type:  PF_COMMENT
        """

        self.enable = PF_ENABLE(conn, *kind)
        """Enable or disable the filter.

        :type:  PF_ENABLE
        """

        self.condition = PF_CONDITION(conn, *kind)
        """Filter condition configuration.

        :type:  PF_CONDITION
        """

    async def delete(self):
        """Delete filter.
        
        :type:  PF_DELETE
        """

        await PF_DELETE(self._conn, *self.kind).set()
        self._observer.notify(idx_obs.IndexEvents.DEL, self)

    @classmethod
    async def _fetch(cls, conn: "itf.IConnection", module_id: int, port_id: int) -> List[int]:
        resp = await PF_INDICES(conn, module_id, port_id).get()
        return list(resp.filter_xindices)

    @classmethod
    async def _new(cls: Type[FT], conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> FT:
        await PF_CREATE(conn, *kind).set()
        return cls(conn, kind, observer)



from __future__ import annotations
from typing import (
    Type,
    TypeVar
)

T = TypeVar("T")


class RegisterModule:
    """Register revision and it's module class"""

    def __init__(self, modules_store: dict[str, Type], rev: str) -> None:
        self.modules_store = modules_store
        self.revision = rev

    def __call__(self, module_type: Type[T]) -> Type[T]:
        if self.revision in self.modules_store:
            raise RuntimeError(f"Module of revision: {self.revision}, is already registered.")
        self.modules_store[self.revision] = module_type
        return module_type



from __future__ import annotations
import asyncio
from typing import Callable
from uuid import uuid4
from .logger import (
    TransportationLogger,
    CustomLogger
)
from .protocol.struct_request import Request
from .protocol.struct_header import ResponseHeader
from ._request_id_counter import RequestIdCounter
from ._stream import StreamReader
from ._processor import PacketsProcessor
from ._publisher import ResponsePublisher
from ._typings import ICommand


class TransportationHandler(asyncio.Protocol):
    """Handling sending and receiving of the XMP commands."""

    __slots__ = (
        "identity",
        "peername",
        "__log",
        "__transport",
        "__id_counter",
        "__stream",
        "__resp_publisher",
        "__pkt_processor"
    )

    def __init__(self, *, enable_logging: bool = False, custom_logger: CustomLogger | None = None) -> None:
        self.identity = uuid4().hex[:6]
        self.peername: tuple[str, int] | None = None
        self.__transport: asyncio.Transport | None = None
        self.__id_counter = RequestIdCounter()
        self.__log = TransportationLogger(
            cid=self.identity,
            enabled=enable_logging,
            logger=custom_logger
        )
        self.__stream = StreamReader(header_struct=ResponseHeader)
        self.__resp_publisher = ResponsePublisher(logger=self.__log)
        self.__pkt_processor = PacketsProcessor(
            stream=self.__stream,
            publish_func=self.__resp_publisher.publish
        )

    @property
    def is_connected(self) -> bool:
        return not (self.__transport is None or self.__transport.is_closing())

    def connection_made(self, transport: asyncio.Transport) -> None:
        self.__transport = transport
        self.peername = transport.get_extra_info("peername")
        self.__pkt_processor.start()
        self.__log.info(f"Connected to {self.peername}")

    def data_received(self, data: bytes) -> None:
        """Process received data from xenaserver."""
        try:
            self.__stream.feed_data(data)
        except AssertionError:
            if self.is_connected:
                self.__stream._eof = False
                self.__stream.feed_data(data)
            else:
                self.close()

    def eof_received(self) -> None:
        self.__stream.feed_eof()
        self.__log.info("EOF received")

    def connection_lost(self, exc: Exception | None) -> None:
        self.__resp_publisher.publish_connection_lost(self.peername)
        self.__transport = None
        self.__pkt_processor.stop()
        if exc:
            self.__log.error(exc)
        else:
            self.__log.info(f"The server {self.peername} closed the connection")

    def send(self, data: bytes | bytearray | memoryview) -> None:
        """
        Send applied commands from sending queue to
        xenaserver and liberate the sending queue.
        """
        if not self.is_connected:
            raise BrokenPipeError("No socket!")
        self.__transport.write(data)  # type: ignore[reportOptionalMemberAccess]

    def close(self) -> None:
        """Close connection with xenaserver."""
        if self.is_connected:
            self.__transport.close()  # type: ignore[reportOptionalMemberAccess]
        self.__transport = None

    async def prepare_data(self, request: Request) -> tuple[bytes, asyncio.Future]:
        assert self.is_connected, "Cannot add command because Socket is disconnected"
        request_id_ = await self.__id_counter.get_number()
        request.update_identifier(request_id_)
        self.__pkt_processor.register(
            req_id=request_id_,
            cmd_code=request.cmd_code
        )
        fut_ = self.__resp_publisher.register_request(
            req_id=request_id_,
            cmd_name=request.class_name
        )
        self.__log.debug_request(request)
        return bytes(request), fut_

    def subscribe(self, xmc_cls: ICommand, callback: Callable) -> None:
        """Register the callback on the command which supports Server PUSH notification."""
        assert xmc_cls.pushed, "Command is not subscribable."
        assert callback, "Callback function is required."
        self.__resp_publisher.subscribe(
            evt=xmc_cls.code,
            func=callback
        )

    def on_disconnected(self, callback: Callable) -> None:
        """Regiser users callback which will be called after connection was terminated."""
        self.__resp_publisher.subscribe_connection_lost(callback)

    def set_outdated(self) -> None:
        pass


import typing
from xoa_driver import ports
from xoa_driver.internals.hli_v1 import revisions
from xoa_driver.internals.utils.managers import ports_manager as pm

if typing.TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from .. import __interfaces as m_itf

from .module_l23_base import ModuleL23
from xoa_driver.internals.commands import (
    M_CLOCKPPBSWEEP,
    M_CLOCKSWEEPSTATUS,
    M_HEALTH,
)


class MClockSweep:
    """Test module local clock sweep"""
    def __init__(self, conn: "itf.IConnection", module_id: int) -> None:
        self.config = M_CLOCKPPBSWEEP(conn, module_id)
        """Configure and control the module local clock sweep.
        Representation of M_CLOCKPPBSWEEP
        """

        self.status = M_CLOCKSWEEPSTATUS(conn, module_id)
        """Status of the module local clock sweep.
        Representation of M_CLOCKSWEEPSTATUS
        """


class ModuleFamilyL1(ModuleL23):
    """Test module Freya family"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)

        self.clock_sweep = MClockSweep(conn, self.module_id)
        """Clock ppm sweep control"""

        self.health = M_HEALTH(conn, self.module_id)
        """Module health info"""


#region Freya 1S 1P
        
@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P[a]")
class MFreya800G1S1P_a(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P[a]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1P_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1P_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P[a]"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P[b]")
class MFreya800G1S1P_b(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1P_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1P_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P[b]"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P-OSFP[a]")
class MFreya800G1S1POSFP_a(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P-OSFP[a]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1POSFP_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1POSFP_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P-OSFP[a]"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P-OSFP[b]")
class MFreya800G1S1POSFP_b(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P-OSFP[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1POSFP_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1POSFP_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P-OSFP[b]"""


#endregion

#region Freya 1S 1P G1
@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P[a] G1")
class MFreya800G1S1P_a_g1(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P[a] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1P_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1P_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P[a] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P[b] G1")
class MFreya800G1S1P_b_g1(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P[b] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1P_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1P_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P[b] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P-OSFP[a] G1")
class MFreya800G1S1POSFP_a_g1(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P-OSFP[a] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1POSFP_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1POSFP_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P-OSFP[a] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P-OSFP[b] G1")
class MFreya800G1S1POSFP_b_g1(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P-OSFP[b] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1POSFP_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1POSFP_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P-OSFP[b] G1"""


#endregion

#region Freya 1S 1P G2
@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P[a] G2")
class MFreya800G1S1P_a_g2(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P[a] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1P_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1P_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P[a] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P[b] G2")
class MFreya800G1S1P_b_g2(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P[b] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1P_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1P_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P[b] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P-OSFP[a] G2")
class MFreya800G1S1POSFP_a_g2(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P-OSFP[a] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1POSFP_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1POSFP_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P-OSFP[a] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-1S-1P-OSFP[b] G2")
class MFreya800G1S1POSFP_b_g2(ModuleFamilyL1):
    """Test module Freya-800G-1S-1P-OSFP[b] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G1S1POSFP_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G1S1POSFP_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-1S-1P-OSFP[b] G2"""


#endregion

#region Freya 1S 4P
@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[a]")
class MFreya800G4S1P_a(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[a]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[a]"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[b]")
class MFreya800G4S1P_b(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[b]"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[c]")
class MFreya800G4S1P_c(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[c]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_c] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_c,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[c]"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[d]")
class MFreya800G4S1P_d(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[d]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_d] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_d,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[d]"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[e]")
class MFreya800G4S1P_e(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[e]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_e] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_e,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[e]"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[f]")
class MFreya800G4S1P_f(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[f]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_f] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_f,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[f]"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[a]")
class MFreya800G4S1POSFP_a(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[a]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[a]"""



@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[b]")
class MFreya800G4S1POSFP_b(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[b]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[b]"""



@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[c]")
class MFreya800G4S1POSFP_c(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[c]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_c] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_c,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[c]"""



@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[d]")
class MFreya800G4S1POSFP_d(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[d]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_d] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_d,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[d]"""



@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[e]")
class MFreya800G4S1POSFP_e(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[e]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_e] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_e,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[e]"""



@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[f]")
class MFreya800G4S1POSFP_f(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[f]"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_f] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_f,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[f]"""


#endregion

#region Freya 1S 4P G1

@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[a] G1")
class MFreya800G4S1P_a_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[a] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[a] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[b] G1")
class MFreya800G4S1P_b_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[b] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[b] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[c] G1")
class MFreya800G4S1P_c_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[c] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_c] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_c,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[c] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[d] G1")
class MFreya800G4S1P_d_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[d] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_d] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_d,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[d] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[e] G1")
class MFreya800G4S1P_e_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[e] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_e] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_e,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[e] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[f] G1")
class MFreya800G4S1P_f_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[f] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_f] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_f,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[f] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[a] G1")
class MFreya800G4S1POSFP_a_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[a] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[a] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[b] G1")
class MFreya800G4S1POSFP_b_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[b] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[b] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[c] G1")
class MFreya800G4S1POSFP_c_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[c] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_c] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_c,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[c] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[d] G1")
class MFreya800G4S1POSFP_d_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[d] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_d] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_d,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[d] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[e] G1")
class MFreya800G4S1POSFP_e_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[e] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_e] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_e,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[e] G1"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[f] G1")
class MFreya800G4S1POSFP_f_g1(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[f] G1"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_f] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_f,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[f] G1"""


#endregion

#region Freya 1S 4P G2        

@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[a] G2")
class MFreya800G4S1P_a_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[a] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[a] G2"""
        

@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[b] G2")
class MFreya800G4S1P_b_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[b] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[b] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[c] G2")
class MFreya800G4S1P_c_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[c] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_c] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_c,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[c] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[d] G2")
class MFreya800G4S1P_d_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[d] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_d] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_d,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[d] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[e] G2")
class MFreya800G4S1P_e_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[e] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_e] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_e,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[e] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P[f] G2")
class MFreya800G4S1P_f_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P[f] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1P_f] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1P_f,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P[f] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[a] G2")
class MFreya800G4S1POSFP_a_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[a] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_a] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_a,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[a] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[b] G2")
class MFreya800G4S1POSFP_b_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[b] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_b] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_b,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[b] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[c] G2")
class MFreya800G4S1POSFP_c_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[c] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_c] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_c,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[c] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[d] G2")
class MFreya800G4S1POSFP_d_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[d] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_d] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_d,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[d] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[e] G2")
class MFreya800G4S1POSFP_e_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[e] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_e] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_e,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[e] G2"""


@typing.final
@revisions.register_valkyrie_module(rev="Freya-800G-4S-1P-OSFP[f] G2")
class MFreya800G4S1POSFP_f_g2(ModuleFamilyL1):
    """Test module Freya-800G-4S-1P-OSFP[f] G2"""
    def __init__(self, conn: "itf.IConnection", init_data: "m_itf.ModuleInitData") -> None:
        super().__init__(conn, init_data)
        self.ports: pm.PortsManager[ports.PFreya800G4S1POSFP_f] = pm.PortsManager(
            conn=conn,
            ports_type=ports.PFreya800G4S1POSFP_f,
            module_id=self.module_id,
            ports_count=self.ports_count
        )
        """Port Index Manager of Freya-800G-4S-1P-OSFP[f] G2"""




#endregion



from typing import (
    TYPE_CHECKING,
    Dict,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from xoa_driver.internals.commands import (
    PR_TPLDJITTER,
    PR_TOTAL,
    PR_NOTPLD,
    PR_EXTRA,
    PR_TPLDS,
    PR_TPLDTRAFFIC,
    PR_TPLDERRORS,
    PR_TPLDLATENCY,
    PR_FILTER,
    PR_CLEAR,
    PR_PFCSTATS,
    # genuine only
    PR_CALIBRATE,
    PR_UAT_STATUS,
    PR_UAT_TIME,

)


# region All Ports

class PrsTPLD:
    """L23 port's TPLD traffic statistics."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, tpld_idx: int) -> None:
        self.traffic = PR_TPLDTRAFFIC(conn, module_id, port_id, tpld_idx)
        """L23 port's statistics of traffic with TPLD.
        Representation of PR_TPLDTRAFFIC
        """

        self.errors = PR_TPLDERRORS(conn, module_id, port_id, tpld_idx)
        """L23 port's error statistics of traffic with TPLD.
        Representation of PR_TPLDERRORS
        """

        self.latency = PR_TPLDLATENCY(conn, module_id, port_id, tpld_idx)
        """L23 port's latency statistics of traffic with TPLD.
        Representation of PR_TPLDLATENCY
        """

        self.jitter = PR_TPLDJITTER(conn, module_id, port_id, tpld_idx)
        """L23 port's jitter statistics of traffic with TPLD.
        Representation of PR_TPLDJITTER
        """


class PortReceptionStatistics:
    """Port RX statistics"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.__conn = conn
        self.__module_id = module_id
        self.__port_id = port_id

        self.__tpld_ids = PR_TPLDS(conn, module_id, port_id)

        self.total = PR_TOTAL(conn, module_id, port_id)
        """L23 port's total traffic statistics.
        Representation of PR_TOTAL
        """

        self.no_tpld = PR_NOTPLD(conn, module_id, port_id)
        """L23 port's statistics of traffic without TPLD.
        Representation of PR_NOTPLD
        """

        self.extra = PR_EXTRA(conn, module_id, port_id)
        """L23 port's extra traffic statistics.
        Representation of PR_EXTRA
        """
        self.pfc_stats = PR_PFCSTATS(conn, module_id, port_id)
        """L23 port's statistics of received PFC packets.
        Representation of PR_PFCSTATS
        """
        self.clear = PR_CLEAR(conn, module_id, port_id)
        """Clear all RX statistics on the L23 port.
        Representation of PR_CLEAR
        """

    def obtain_filter_statistics(self, filter: int) -> "PR_FILTER":
        """Obtain L23 port filtered traffic statistics.

        :param filter: index of the filter
        :type filter: int
        :return: filtered traffic statistics
        :rtype: PR_FILTER
        """

        return PR_FILTER(
            self.__conn,
            self.__module_id,
            self.__port_id,
            filter
        )

    def access_tpld(self, tpld_id: int) -> "PrsTPLD":
        """Access a certain TPLD value on a L23 port.

        :param tpld_id: TPLD value.
        :type tpld_id: int
        :return: L23 port's TPLD traffic statistics
        :rtype: PrsTPLD
        """
        return PrsTPLD(self.__conn, self.__module_id, self.__port_id, tpld_id)

    async def obtain_available_tplds(self) -> Dict[int, "PrsTPLD"]:
        #TODO: need to check use case if this is enough, otherwise to extend by give possibility manually specify indices
        tplds = await self.__tpld_ids.get()
        return {
            idx:
            PrsTPLD(self.__conn, self.__module_id, self.__port_id, idx)
            for idx in tplds.test_payload_identifiers
        }


# endregion

# region Genuine Ports
class GPrsUat():
    """L23 port UAT info."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.status = PR_UAT_STATUS(conn, module_id, port_id)
        """L23 port UAT status.
        Representation of PR_UAT_STATUS
        """

        self.time = PR_UAT_TIME(conn, module_id, port_id)
        """L23 port number of unavailable seconds.
        Representation of PR_UAT_TIME
        """


class GenuinePortReceptionStatistics(PortReceptionStatistics):
    """L23 port RX statistics."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.calibrate = PR_CALIBRATE(conn, module_id, port_id)
        """L23 port calibration of latency for received packets.
        Representation of PR_CALIBRATE
        """

        self.uat = GPrsUat(conn, module_id, port_id)
        """L23 port UAT info.
        """

# endregion




import functools
from typing import TYPE_CHECKING, Tuple
from typing_extensions import Self
from xoa_driver.internals.commands import (
    P_DYNAMIC,
    # PL1_AUTONEGINFO,
    # PL1_LINKTRAININFO,
    # PL1_LOG,
    # PL1_CFG_TMP,
)
from xoa_driver.internals.utils import attributes as utils
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf

from .bases.port_l23_genuine import BasePortL23Genuine
from .pcs_pma_ijkl_chimera import PcsPma as PcsPma1
from .pcs_pma_ghijkl import (
    PcsPma as PcsPma2,
    SerDes,
)
from .pcs_pma_l import PcsPmaL1 as PcsPma3  # PcsPmaL1 for current version is different if compare to L family


class PcsPma(PcsPma1, PcsPma2, PcsPma3):
    def __init__(self, conn: "itf.IConnection", port) -> None:
        PcsPma1.__init__(self, conn, port)
        PcsPma2.__init__(self, conn, port)
        PcsPma3.__init__(self, conn, port)


class FamilyL1(BasePortL23Genuine):
    pcs_pma: PcsPma
    serdes: Tuple[SerDes, ...]

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        super().__init__(conn, module_id, port_id)
        self.dynamic = P_DYNAMIC(conn, module_id, port_id)
        """L23 port's dynamic traffic change.
        Representation of P_DYNAMIC
        """

        # self.fault = Fault(conn, module_id, port_id)

    async def _setup(self) -> Self:
        await super()._setup()
        self.pcs_pma = PcsPma(self._conn, self)
        self.serdes = tuple(
            SerDes(self._conn, *self.kind, serdes_xindex=serdes_xindex)
            for serdes_xindex in range(self.info.capabilities.serdes_count)
        )
        return self

    on_dynamic_change = functools.partialmethod(utils.on_event, P_DYNAMIC)
    """Register a callback to the event that the port's dynamic traffic setting changes."""


class PFreya800G1S1P_a(FamilyL1):
    """L23 port on Freya-800G-1S-1P[a] module.
    """
    ...


class PFreya800G1S1P_b(FamilyL1):
    """L23 port on Freya-800G-1S-1P[b] module.
    """
    ...


class PFreya800G1S1POSFP_a(FamilyL1):
    """L23 port on Freya-800G-1S-1P-OSFP[a] module.
    """
    ...

class PFreya800G1S1POSFP_b(FamilyL1):
    """L23 port on Freya-800G-1S-1P-OSFP[b] module.
    """
    ...

class PFreya800G4S1P_a(FamilyL1):
    """L23 port on Freya-800G-4S-1P[a] module.
    """
    ...


class PFreya800G4S1P_b(FamilyL1):
    """L23 port on Freya-800G-4S-1P[b] module.
    """
    ...


class PFreya800G4S1P_c(FamilyL1):
    """L23 port on Freya-800G-4S-1P[c] module.
    """
    ...


class PFreya800G4S1P_d(FamilyL1):
    """L23 port on Freya-800G-4S-1P[d] module.
    """
    ...


class PFreya800G4S1P_e(FamilyL1):
    """L23 port on Freya-800G-4S-1P[e] module.
    """
    ...


class PFreya800G4S1P_f(FamilyL1):
    """L23 port on Freya-800G-4S-1P[f] module.
    """
    ...


class PFreya800G4S1POSFP_a(FamilyL1):
    """L23 port on Freya-800G-4S-1P-OSFP[a] module.
    """
    ...


class PFreya800G4S1POSFP_b(FamilyL1):
    """L23 port on Freya-800G-4S-1P-OSFP[b] module.
    """
    ...


class PFreya800G4S1POSFP_c(FamilyL1):
    """L23 port on Freya-800G-4S-1P-OSFP[c] module.
    """
    ...


class PFreya800G4S1POSFP_d(FamilyL1):
    """L23 port on Freya-800G-4S-1P-OSFP[d] module.
    """
    ...


class PFreya800G4S1POSFP_e(FamilyL1):
    """L23 port on Freya-800G-4S-1P-OSFP[e] module.
    """
    ...


class PFreya800G4S1POSFP_f(FamilyL1):
    """L23 port on Freya-800G-4S-1P-OSFP[f] module.
    """
    ...


from __future__ import annotations
import asyncio
from asyncio.events import AbstractEventLoop
import io
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
)
if TYPE_CHECKING:
    from .transporter.handler import TransportationHandler
    from . import interfaces

from .token import Token
from .transporter import exceptions


async def establish_connection(transporter: "TransportationHandler", host: str, port: int = 22606, *, seconds_timeout: int = 5, loop: AbstractEventLoop | None = None) -> None:
    """
    Establish connection to provided host and port and assign ``<TransportationHandler>`` to it.
    """
    assert isinstance(loop, AbstractEventLoop) or loop is None, "<loop> must be an instance of AbstractEventLoop or None"
    __loop = loop if loop else asyncio.get_event_loop()
    try:
        await asyncio.wait_for(
            __loop.create_connection(lambda: transporter, host=host, port=port),
            seconds_timeout
        )
    except OSError:
        raise exceptions.XoaConnectionError(host, port) from None
    except asyncio.exceptions.TimeoutError:
        raise exceptions.XoaConnectionTimeoutError(host, port, seconds_timeout) from None


async def apply_iter(*cmd_tokens: Token[Any], return_exceptions: bool = False, token_timeout_sec: float | None = 5.0) -> AsyncGenerator[Any, None]:
    """
    Main interface for chunking the commands which need to be send to the single tester at the same time.
    """
    if not cmd_tokens:
        return
    conn: "interfaces.IConnection" = cmd_tokens[0].connection
    buffer_bytes = io.BytesIO()
    queue: asyncio.Queue[asyncio.Future] = asyncio.Queue()
    for t in cmd_tokens:
        (data, fut) = await t.connection.prepare_data(t.request)
        buffer_bytes.write(data)
        queue.put_nowait(fut)
    conn.send(buffer_bytes.getvalue())
    buffer_bytes.close()
    del buffer_bytes

    while not queue.empty():
        future = await queue.get()
        try:
            result_ = await asyncio.wait_for(
                asyncio.shield(future),
                token_timeout_sec if return_exceptions else None
            )
        except Exception as e:
            if return_exceptions:
                yield e
            else:
                raise e
        else:
            yield result_
        queue.task_done()
    await queue.join()


async def apply(*cmd_tokens: Token[Any], return_exceptions: bool = False, token_timeout_sec: float | None = 5.0) -> list[Any]:
    """
    Main interface for chunking the commands which need to be send to one or multiple testers at the same time.
    """
    assert len(cmd_tokens) <= 200, "Number of the commands is bigger then 200 for one aggregation, please use function <apply_iter> instead"
    return [
        f async for f in apply_iter(
            *cmd_tokens, 
            return_exceptions=return_exceptions, 
            token_timeout_sec=token_timeout_sec
        )
    ]



from typing import (
    final,
    List,
    Type,
    TypeVar,
    TYPE_CHECKING,
)

from xoa_driver.internals.commands import (
    P4G_INDICES,
    P4G_CREATE,
    P4G_DELETE,
    P4G_ENABLE,
    P4G_COMMENT,
    P4G_CLEAR_COUNTERS,
    P4G_ROLE,
    P4G_LP_TIME_SCALE,
    P4G_LP_SHAPE,
    P4G_TEST_APPLICATION,
    P4G_L4_PROTOCOL,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
    from xoa_driver.internals.utils import kind
from xoa_driver.internals.utils.indices import observer as idx_obs
from .tls import GTls
from .l2 import GL2
from .raw import GRaw
from .tcp import GTcp
from .udp import GUdp
from .replay import GReplay
from .l3 import GL3
from .user_state import GUserState
from .histogram import GHistogram

from ..base_index import BaseIndex


class GCounters:
    """Connection Group's counters
    """
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.clear = P4G_CLEAR_COUNTERS(conn, module_id, port_id, group_idx)
        """Clear the counters
        
        :type: P4G_CLEAR_COUNTERS
        """


class GLoadProfile:
    """Connection Group's load profile.
    """
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, group_idx: int) -> None:
        self.time_scale = P4G_LP_TIME_SCALE(conn, module_id, port_id, group_idx)
        """Time scale of the load profile
        
        :type: P4G_LP_TIME_SCALE
        """

        self.shape = P4G_LP_SHAPE(conn, module_id, port_id, group_idx)
        """Shape of the load profile.
        
        :type: P4G_LP_SHAPE
        """


CG = TypeVar("CG")


@final
class ConnectionGroupIdx(BaseIndex):
    """L47 Connection Group Index Manager"""
    def __init__(self, conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> None:
        super().__init__(conn, kind, observer)

        self.comment = P4G_COMMENT(self._conn, *kind)
        """Description of the connection group
        
        :type: P4G_COMMENT
        """

        self.status = P4G_ENABLE(self._conn, *kind)
        """Enable/disable/suppress a previously created connection group
        
        :type: P4G_ENABLE
        """

        self.role = P4G_ROLE(self._conn, *kind)
        """Specifies the client or server role for this Connection Group.
        
        :type: P4G_ROLE
        """

        self.layer4_protocol = P4G_L4_PROTOCOL(self._conn, *kind)
        """Specifies either TCP or UDP as Layer 4 protocol.
        
        :type: P4G_L4_PROTOCOL
        """

        self.test_application = P4G_TEST_APPLICATION(self._conn, *kind)
        """Configure the application layer mode.
        
        :type: P4G_TEST_APPLICATION
        """

        self.tls = GTls(self._conn, *kind)
        """TLS configurations.
        
        :type: GTls
        """

        self.l2 = GL2(self._conn, *kind)
        """L2 configurations.
        
        :type: GL2
        """

        self.raw = GRaw(self._conn, *kind)
        """Raw configurations.
        
        :type: GRaw
        """

        self.tcp = GTcp(self._conn, *kind)
        """TCP configurations.
        
        :type: GTcp
        """

        self.udp = GUdp(self._conn, *kind)
        """UDP configurations.
        
        :type: GUdp
        """

        self.replay = GReplay(self._conn, *kind)
        """Replay configurations.
        
        :type: GReplay
        """

        self.l3 = GL3(self._conn, *kind)
        """L3 configurations.
        
        :type: GL3
        """

        self.user_state = GUserState(self._conn, *kind)
        """User state configurations.
        
        ;type: GUserState
        """

        self.histogram = GHistogram(self._conn, *kind)
        """Histogram configurations.
        
        :type: GHistogram
        """

        self.counters = GCounters(self._conn, *kind)
        """Counters.
        
        :type: GCounters
        """

        self.load_profile = GLoadProfile(self._conn, *kind)
        """Load Profile configurations.
        
        :type: GLoadProfile
        """

    async def delete(self):
        """Delete the Connection Group.
        """
        await P4G_DELETE(self._conn, *self.kind).set()
        self._observer.notify(idx_obs.IndexEvents.DEL, self)

    @classmethod
    async def _fetch(cls, conn: "itf.IConnection", module_id: int, port_id: int) -> List[int]:
        resp = await P4G_INDICES(conn, module_id, port_id).get()
        return list(resp.group_identifiers)

    @classmethod
    async def _new(cls: Type[CG], conn: "itf.IConnection", kind: "kind.IndicesKind", observer: "idx_obs.IndicesObserver") -> CG:
        await P4G_CREATE(conn, *kind).set()
        return cls(conn, kind, observer)



from typing import (
    TYPE_CHECKING,
    List,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    PEC_INDICES,
    PEC_VAL,
    PEC_COMMENT,
    PEC_DELETE,
    PEC_DISTTYPE,
)

from xoa_driver.internals.utils.indices import observer


class CustomDistribution:
    """Custom distribution"""

    def __init__(self, observer: "observer.IndicesObserver", conn: "itf.IConnection", module_id: int, port_id: int, custom_distribution_index: int) -> None:
        self.__observer = observer
        self.__conn = conn
        self.__module_id = module_id
        self.__port_id = port_id
        self.__cdi = custom_distribution_index
        self.definition = PEC_VAL(conn, module_id, port_id, custom_distribution_index)
        """Custom distribution definition.

        :type: PEC_VAL
        """

        self.comment = PEC_COMMENT(conn, module_id, port_id, custom_distribution_index)
        """Custom distribution description.

        :type: PEC_COMMENT
        """

        self.type = PEC_DISTTYPE(conn, module_id, port_id, custom_distribution_index)
        """Custom distribution type.

        :type: PEC_DISTTYPE
        """

    async def delete(self) -> None:
        """
        Deleting an existing Custom Distribution
        """

        await PEC_DELETE(
            self.__conn,
            self.__module_id,
            self.__port_id,
            self.__cdi
        ).set()
        self.__observer.notify(observer.IndexEvents.DEL, self)


class CustomDistributions:
    """Custom distributions"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.__conn = conn
        self.__module_id = module_id
        self.__port_id = port_id
        self.__items: List[CustomDistribution] = []
        self.__observer = observer.IndicesObserver()
        self.__observer.subscribe(
            observer.IndexEvents.DEL,
            self.__remove_from_slot
        )

    async def server_sync(self) -> None:
        """Sync the indices with xenaserver"""

        _resp = await PEC_INDICES(self.__conn, self.__module_id, self.__port_id).get()
        self.__items = [
            CustomDistribution(
                self.__observer,
                self.__conn,
                self.__module_id,
                self.__port_id,
                idx
            )
            for idx in _resp.indexations
        ]

    def __len__(self) -> int:
        """Return the number of existing indices"""
        return len(self.__items)

    def __iter__(self):
        self.__k = 0
        return self

    def __next__(self):
        try:
            v = self.__items[self.__k]
        except IndexError:
            raise StopIteration()
        else:
            self.__k += 1
            return v

    def __getitem__(self, key: int):
        return self.__items[key]

    def __remove_from_slot(self, index_inst: "CustomDistribution") -> None:
        # throws ValueError if element is not exists in list of indices
        self.__items.remove(index_inst)

    async def assign(self, idx_cuantity: int = 0) -> None:
        """
        Assign Custom distribution indices, all indices which is out of range will be removed.
        ``idx_cuantity`` permitted values is: 0 <= idx_cuantity <= 40
        """

        if not (0 <= idx_cuantity <= 40):
            raise ValueError("idx_cuantity must be in range of: 0 <= idx_cuantity <= 40")
        await PEC_INDICES(self.__conn, self.__module_id, self.__port_id).set([i for i in range(idx_cuantity)])
        await self.server_sync()

    async def remove(self, position_idx: int) -> None:
        """Remove a index from port"""
        await self.__items[position_idx].delete()
